////////////////////////////////////////////////////////////////////////////
//
// Voltmeter and Ammeter
//
// Copyright by rubidesu 2023
//
// http://www.gatelink.co.jp/hw/
// Twitter : @rubidesu
//
////////////////////////////////////////////////////////////////////////////

//#define _CLOCK_32MHZ_
//#define _CLOCK_8MHZ_
#define _CLOCK_2MHZ_

#define _I2C_LCD_
//#undef _I2C_LCD_

//#define _HW_USART_
#undef _HW_USART_

#include <xc.h>

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (HS Oscillator, High-speed crystal/resonator connected between OSC1 and OSC2 pins)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#ifdef _CLOCK_32MHZ_
	#pragma config PLLEN = ON       // PLL Enable (4x PLL disabled)
#else //_CLOCK_32MHZ_
	#pragma config PLLEN = OFF      // PLL Enable (4x PLL disabled)
#endif //_CLOCK_32MHZ_
#pragma config STVREN = OFF     // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will not cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)

#if defined _CLOCK_32MHZ_
	#define _XTAL_FREQ 32000000
#elif defined _CLOCK_8MHZ_
	#define _XTAL_FREQ  8000000
#elif defined _CLOCK_2MHZ_
	#define _XTAL_FREQ  2000000
#else
	#error Clock Error.
#endif

//^`.
typedef unsigned char	BYTE;
typedef          bit	BIT;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;

//High/Low̒`.
#define HIGH 1
#define LOW  0

#ifndef _HW_USART_
	#define SERIAL_TX	LATAbits.LATA5		//TXs.

	//#define SERIAL_WAIT	104			//9600bps {̒l.
	#define SERIAL_WAIT	95			//9600bps 2MHz l.
#endif //_HW_USART_

#ifdef _I2C_LCD_
	#define IIC_WRITE	0x00
	#define IIC_READ	0x01

	#define IIC_ACK		0x00
	#define IIC_NACK	0x01

	#define I2C_SCL			LATAbits.LATA0
	#define I2C_SDA_IN		RA1
	#define I2C_SDA			LATAbits.LATA1
	#define I2C_SDA_TRIS	TRISAbits.TRISA1

	#define RS_DATA			0x40	//LCDf[^.
	#define RS_CMND			0x00	//LCDR}h.

	#define ADDR_LCD		0x7c	//LCD̃AhX.

#endif //_I2C_LCD_

#define UP_VOL			3.3
#define UP_AD_RESOL		1024.0
#define VOL_DEVIDE		10.0
#define CUR_DEVIDE		10.0

#define AD_VOLTAGE		0b00010		//AN2.
#define AD_CURRENT		0b00011		//AN3.

typedef union{
	BYTE bByte;
	struct{
		unsigned b7 : 1;
		unsigned b6 : 1;
		unsigned b5 : 1;
		unsigned b4 : 1;
		unsigned b3 : 1;
		unsigned b2 : 1;
		unsigned b1 : 1;
		unsigned b0 : 1;
	} bit_t;
} s_reg_t;
s_reg_t s_reg;

BYTE bLcdStr[5];

#define AD_AVERAGE_MAX	10

////////////////////////////////////////////////////////////////////////////
void chip_initialize(void)
{
	WORD wBaud;

	//NbNݒ.

#if defined _CLOCK_32MHZ_
	OSCCON = 0b01110000;		//32MHz.
#elif defined _CLOCK_8MHZ_
	OSCCON = 0b01110010;		//8MHz.
#elif defined _CLOCK_2MHZ_
	OSCCON = 0b01100010;		//2MHz.
#else
	#error Clock Error.
#endif

	//vAbvݒ.
	OPTION_REGbits.nWPUEN = 1;	//0:vAbvL 1:.

	//|[g̓o͐ݒ.
	TRISA = 0b00011100;	//0:o 1:.

	//Rp[^ݒ.
	CM1CON0 = 0b00000000;	//ׂăfW^.

	//|[g̗prݒ.
	ANSELA = 0b00010100;		//A/Dϊ|[g=1.

	ADCON0 = 0b00001000;		//AN2=AiO.
	ADCON1 = 0b10100000;		//ADʂEl. dVcc ϊNbN/32.

	ADCON0bits.ADON = 1;	//Start A/D converter.

#ifdef _HW_USART_
	//USART̐ݒ/////////////////////////////////
	APFCONbits.TXCKSEL = 1;		//VAMRA4.
	APFCONbits.RXDTSEL = 1;		//VAMRA5.
	wBaud = 51;					//9600bps 2MHz.

	TXSTA = 0b00100100;			//BRGH = 1.
	RCSTA = 0b10010000;
	BAUDCONbits.BRG16 = 1;
	BAUDCONbits.SCKP = RA3;
	SPBRGH = (BYTE)(wBaud >> 8);
	SPBRGL = (BYTE)(wBaud & 0xff);
	RCIF = 0;
#else //_HW_USART_
	SERIAL_TX = LOW;
#endif //_HW_USART_
}

#ifdef _I2C_LCD_
////////////////////////////////////////////////////////////////////////////
void i2c_start(void)
{
	I2C_SDA = LOW;
	I2C_SDA_TRIS = 0;
}

////////////////////////////////////////////////////////////////////////////
void i2c_stop(void)
{
	I2C_SCL = LOW;
	__delay_us(1);

	I2C_SDA = LOW;

	I2C_SDA_TRIS = 0;

	I2C_SCL = HIGH;
	__delay_us(1);

	I2C_SDA = HIGH;
	__delay_us(1);
}

////////////////////////////////////////////////////////////////////////////
BYTE i2c_write(BYTE bDat)
{
	BYTE bAckNack = IIC_NACK;

	I2C_SDA_TRIS = 0;

	for(int i = 0; i < 8; i++){
		I2C_SCL = LOW;

		if((bDat & 0x80) == 0x80){
			I2C_SDA = HIGH;
		}
		else{
			I2C_SDA = LOW;
		}

		I2C_SCL = HIGH;

		bDat <<=1;
	}

	//Ackp̃NbN.
	I2C_SCL = LOW;

	I2C_SDA_TRIS = 1;
	__delay_us(2);

	I2C_SCL = HIGH;
	__delay_us(2);

	bAckNack = I2C_SDA_IN;

	return bAckNack;
}

////////////////////////////////////////////////////////////////////////////
BYTE i2c_read(BYTE boAckFlg)
{
	BYTE bDat = 0x00;

	for(BYTE i = 0; i < 8; i++){
		I2C_SCL = LOW;

		I2C_SDA_TRIS = 1;
		__delay_us(3);

		I2C_SCL = HIGH;

		bDat <<= 1;
		bDat |= I2C_SDA_IN;
	}

	I2C_SCL = LOW;

	if(boAckFlg == IIC_ACK){
		I2C_SDA = HIGH;
	}
	else{
		I2C_SDA = LOW;
	}

	I2C_SDA_TRIS = 0;
	__delay_us(2);

	I2C_SCL = HIGH;

	return bDat;
}

////////////////////////////////////////////////////////////////////////////
void i2c_lcd_write(BYTE rs, BYTE dat)
{
	i2c_start();
	i2c_write(ADDR_LCD | IIC_WRITE);	//address
	i2c_write(0x80 | rs);
	i2c_write(dat);
	i2c_stop();
}

////////////////////////////////////////////////////////////////////////////
void i2c_lcd_clear(void)
{
	i2c_lcd_write(RS_CMND, 0x01);
	__delay_ms(2);
}

////////////////////////////////////////////////////////////////////////////
void i2c_lcd_putch(char c)
{
	i2c_lcd_write(RS_DATA, c);
}

////////////////////////////////////////////////////////////////////////////
void i2c_lcd_goto(BYTE pos)
{
	i2c_lcd_write(RS_CMND, 0x80+pos);
}

////////////////////////////////////////////////////////////////////////////
void i2c_lcd_init(void)
{
	BYTE bContVal = 25;

	i2c_lcd_write(RS_CMND, 0b00111000);	//0x38 function set.
	i2c_lcd_write(RS_CMND, 0b00111001);	//0x39 function set.
	i2c_lcd_write(RS_CMND, 0b00010100);	//0x14 internal osc frequency.
	i2c_lcd_write(RS_CMND, 0x70 | (bContVal & 0x0f) /*0b01110000*/);	//0x70 contrast set.
	i2c_lcd_write(RS_CMND, 0x54 | (bContVal >> 4) /*0b01010100*/);	//0x56 power/icon/contrast control.
	i2c_lcd_write(RS_CMND, 0b01101100);	//0x6C follower control.
	__delay_ms(210);
	i2c_lcd_write(RS_CMND, 0b00111000);	//0x38 function set.
	i2c_lcd_write(RS_CMND, 0b00001100);	//0x0C display on/off control.
	i2c_lcd_write(RS_CMND, 0b00000001);	//0x01 clear display.
	__delay_ms(2);
}

////////////////////////////////////////////////////////////////////////////
void write_format_val(void)
{
	i2c_lcd_putch(bLcdStr[0]);
	i2c_lcd_putch(bLcdStr[1]);
	i2c_lcd_putch(bLcdStr[2]);
	i2c_lcd_putch('.');
	i2c_lcd_putch(bLcdStr[3]);
	i2c_lcd_putch(bLcdStr[4]);
}
#endif //_I2C_LCD_

////////////////////////////////////////////////////////////////////////////
void usart_putc(BYTE bDat)
{
#ifdef _HW_USART_
	while(TXSTAbits.TRMT != 1);

	TXREG = bDat;
#else //_HW_USART_
	s_reg.bByte = bDat ^ 0xff;

	//X^[grbg.
	SERIAL_TX = HIGH;
	__delay_us(SERIAL_WAIT);

	SERIAL_TX = s_reg.bit_t.b7;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b6;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b5;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b4;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b3;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b2;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b1;
	__delay_us(SERIAL_WAIT);
	SERIAL_TX = s_reg.bit_t.b0;
	__delay_us(SERIAL_WAIT);

	//Xgbvrbg.
	SERIAL_TX = LOW;
	__delay_us(SERIAL_WAIT);
#endif //_HW_USART_
}

////////////////////////////////////////////////////////////////////////////
void send_num4(WORD val)
{
	usart_putc((val / 1000) + '0');
	val %= 1000;

	usart_putc((val / 100) + '0');
	val %= 100;

	usart_putc((val / 10) + '0');
	val %= 10;

	usart_putc(val + '0');
}

////////////////////////////////////////////////////////////////////////////
void send_format_val(void)
{
	usart_putc(bLcdStr[0]);
	usart_putc(bLcdStr[1]);
	usart_putc(bLcdStr[2]);
	usart_putc('.');
	usart_putc(bLcdStr[3]);
	usart_putc(bLcdStr[4]);
}

////////////////////////////////////////////////////////////////////////////
void format_val(WORD val)
{
	bLcdStr[0] = val / 10000 + '0';
	val %= 10000;

	bLcdStr[1] = val / 1000 + '0';
	val %= 1000;

	bLcdStr[2] = val / 100 + '0';
	val %= 100;

	bLcdStr[3] = val / 10 + '0';
	val %= 10;

	bLcdStr[4] = val + '0';


	if(bLcdStr[0] == '0'){
		bLcdStr[0] = ' ';
	}

	if(bLcdStr[0] == ' ' && bLcdStr[1] == '0'){
		bLcdStr[1] = ' ';
	}
}

////////////////////////////////////////////////////////////////////////////
WORD get_adval(BYTE bCh)
{
	WORD wADVal;

	ADCON0bits.CHS = bCh;
	__delay_us(10);

	ADCON0bits.GO = 1;
	while(ADCON0bits.GO == 1);

	wADVal = ADRES;

	return wADVal;
}

////////////////////////////////////////////////////////////////////////////
void main(void)
{
	WORD wADVol;
	WORD wVol;
	WORD wADCur;
	WORD wCur;
	BYTE i;

	chip_initialize();

	__delay_ms(100);

#ifdef _I2C_LCD_
	i2c_lcd_init();
	i2c_lcd_clear();
#endif //_I2C_LCD_

	while(1){
		//dlvĕω ////////
		for(i = 0, wADVol = 0; i < AD_AVERAGE_MAX; i++){
			wADVol += get_adval(AD_VOLTAGE) & 0x3ff;
			__delay_ms(5);
		}
		wADVol /= AD_AVERAGE_MAX;
		wVol = (WORD)((UP_VOL / UP_AD_RESOL) * (double)wADVol * VOL_DEVIDE * 100.0);

		//dlvĕω ////////
		for(i = 0, wADCur = 0; i < AD_AVERAGE_MAX; i++){
			wADCur += get_adval(AD_CURRENT) & 0x3ff;
			__delay_ms(5);
		}
		wADCur /= AD_AVERAGE_MAX;
		wCur = (WORD)((UP_VOL / UP_AD_RESOL) * (double)wADCur * (100000.0 / CUR_DEVIDE));

		//dl/////////////////////////
		format_val(wVol);

		i2c_lcd_goto(0x00);
		write_format_val();
		i2c_lcd_putch(' ');
		i2c_lcd_putch('V');

		usart_putc(' ');
		send_format_val();
		usart_putc('V');
		usart_putc(':');
		send_num4(wADVol);
		usart_putc('/');

		//dl/////////////////////////
		format_val(wCur);

		usart_putc(' ');
		send_format_val();
		usart_putc('m');
		usart_putc('A');
		usart_putc(':');
		send_num4(wADCur);
		usart_putc('\r');
		usart_putc('\n');

		i2c_lcd_goto(0x40);
		write_format_val();
		i2c_lcd_putch('m');
		i2c_lcd_putch('A');

		__delay_ms(150);		//3Hz.
	}
}
